/*
 * Copyright 2025 Naomi (boyfailure.dev).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dev.boyfailure.vectorbreakout;

/**
 * The balls that move around the playfield and hit bricks.
 * @author Naomi (boyfailure.dev)
 * @since 20250323
 */
public class Ball extends Entity {

    private boolean isActive = false;
    private boolean isInvincible = false;
    byte ballColor = 0;
    
    byte speedX = 5;
    byte speedY = 5;

    private GameState gameState;

    public Ball(GameState gameState, int x, int y, int width, int height, int color, boolean active) {
        this.gameState = gameState;
        this.posX = (short) x;
        this.posY = (short) y;
        this.width = (short) width;
        this.height = (short) height;
        this.ballColor = (byte) color;
        this.hitBox = new Hitbox(x, y, width, height, 1);
        this.isActive = active;
    }

    public void spawnBall(boolean useLFSR) {
        if (useLFSR) {this.spawnBall(((gameState.getPokey().getDecimalValue() / 2) + gameState.getPokey().getShiftedDecimalValue() + 30), (this.posY + (gameState.getPokey().getDecimalValue() / 4)) - 40);}
        else {this.spawnBall(this.posX, this.posY);}
    }

    public void spawnBall(int x, int y) {
        this.posX = (short) x;
        this.posY = (short) y;
        this.isActive = true;
        if (gameState.getBalls().size() <= 1) {
            gameState.removeLife();
            gameState.getTextElements().get(4).setText(String.valueOf(gameState.getLives()));
        }
    }

    public void moveBall() {
        this.posX += this.speedX;
        this.posY += this.speedY;

        this.hitBox.moveTo(this.posX, this.posY, this.width, this.height);

        this.checkCollision();
    }

    /**
     * Richochets the ball.
     * @param bounceType The code for which direction the ball bounces
     */
    public void bounce(int bounceType) {
        switch((byte) bounceType) {
            case 0:
                this.speedY = (byte) -(this.speedY);
                byte randByte = (byte) (gameState.getPokey().getDecimalValue() / 16);
                if (randByte < 0) {randByte = 0;}
                else if (randByte > 15) {randByte = 15;}
                switch (randByte) {
                    case 0: break;
                    case 1: this.speedX++; break;
                    case 2: break;
                    case 3: this.speedX--; break;
                    case 4: break;
                    case 5: this.speedY++;break;
                    case 6: break;
                    case 7: this.speedY--;break;
                    case 8: break;
                    case 9: this.speedY++;speedX++;break;
                    case 10: break;
                    case 11: this.speedY--;speedX--;break;
                    case 12: break;
                    case 13: this.speedY++;speedY++;break;
                    case 14: break;
                    case 15: this.speedY--;speedY--;break;
                }
                break;
            case 1:
                this.speedX = (byte) -(this.speedX);
                break;
            case 2:
                if (this.speedX > 0) {
                    this.speedX = (byte) -(this.speedX);
                }
                else {
                    this.speedX -= 4;
                }
                this.speedY = (byte) -(this.speedY);
                break;
            case 3:
                if (this.speedX < 0) {
                    this.speedX = (byte) -(this.speedX);
                }
                else {
                    this.speedX += 4;
                }
                this.speedY = (byte) -(this.speedY);
                break;
        }
        if (this.speedX < 2 && this.speedX >= 0) {this.speedX = 2;}
        else if (this.speedX > -2 && this.speedX < 0) {this.speedX = -2;}
        else if (this.speedX > 11) {this.speedX -= 3;}  // Slow down if too fast
        if (this.speedY < 4 && this.speedY >= 0) {this.speedY = 4;}
        else if (this.speedY > -4 && this.speedY < 0) {this.speedY = -4;}
        else if (this.speedY > 11) {this.speedY -= 3;} // Slow down if too fast
    }

    public void destroyBall() {
        this.isActive = false;
        this.posX = 32767;
        this.posY = 32767;
        this.hitBox.setBounds(32766, 32766, 32767, 32767);
        if (gameState.getBalls().size() <= 1) {
            if (gameState.getLives() <= 0) {
                gameState.gameLose();
            }
        }
    }

    public void cull() {
        this.isActive = false;
        this.posX = 32767;
        this.posY = 32767;
        this.hitBox.setBounds(32766, 32766, 32767, 32767);
    }

    /**
     * Checks to see if the ball is in contact with other Entities and walls.
     */
    public synchronized void checkCollision() {

        this.setInvincibilityState(true); // Delay bouncing again for a frame

        // Check floor
        if (this.posY >= gameState.getInternalResY()) {
            this.destroyBall();
            return;
        }
        // Check ceiling
        else if (this.posY <= 0) {
            this.bounce(0);
            this.posY = 0;
            this.wallBounceAnimation();
            return;
        }

        // Check walls
        if (this.posX <= 20) {
            this.bounce(1);
            this.posX = 22;
            this.wallBounceAnimation();
            return;
        }
        else if ((this.posX + this.width) >= (gameState.getInternalResX() - 20 - this.width)) {
            this.bounce(1);
            this.posX = (short) (gameState.getInternalResX() - 25 - this.width);
            this.wallBounceAnimation();
            return;
        }

        // Check bricks
        if (gameState.getCurrentLevel() != null) {
            if (gameState.getCurrentLevel().brickList != null) {
                for (Brick brick : gameState.getCurrentLevel().brickList) {
                    // Check for left bounce
                    if (this.hitBox.collides(brick.hitBoxLeft.getBounds())) {
                        if (this.hitBox.collides(brick.hitBox.getBounds())) {
                            this.bounce(0);
                            this.bounce(1);
                        }
                        else {
                            this.bounce(1);
                        }
                        brick.breakBrick();
                        return;
                    }
                    // Check for right bounce
                    else if (this.hitBox.collides(brick.hitBoxRight.getBounds())) {
                        if (this.hitBox.collides(brick.hitBox.getBounds())) {
                            this.bounce(0);
                            this.bounce(1);
                        }
                        else {
                            this.bounce(1);
                        }
                        brick.breakBrick();
                        return;
                    }
                    // Check for vertical bounce
                    else if (this.hitBox.collides(brick.hitBox.getBounds())) {
                        brick.breakBrick();
                        this.bounce(0);
                        return;
                    }
                }
            }
        }

        // Check paddles
        for (Paddle paddle : gameState.getPaddles()) {
            if (this.hitBox.collides(paddle.hitBox.getBounds())) {
                if (this.hitBox.collides(paddle.hitBoxLeft.getBounds())) {
                    this.bounce(2);
                }
                else if (this.hitBox.collides(paddle.hitBoxRight.getBounds())) {
                    this.bounce(3);
                }
                else {
                    this.bounce(0);
                }

                this.posY -= 2;

                // Draw particles
                Particle particle1, particle2, particle3;
                particle1 = new Particle(gameState, 0, this.posX, this.posY - 10, 0, -2, 1, 2, 7, 3);
                particle2 = new Particle(gameState, 0, this.posX - 3, this.posY - 10, -2, -4, 1, 2, 6, 3);
                particle3 = new Particle(gameState, 0, this.posX + 10, this.posY - 10, 2, -4, 1, 2, 8, 3);
                gameState.getParticles().add(particle1);
                gameState.getParticles().add(particle2);
                gameState.getParticles().add(particle3);
                particle1.spawn(5);
                particle2.spawn(5);
                particle3.spawn(5);

                return; // Prevents bouncing multiple times
            }
        }
    }

    private void wallBounceAnimation() {
        Particle particle1, particle2, particle3, particle4;
        particle1 = new Particle(gameState, 0, this.posX - 3, this.posY - 10, -2, -2, 1, 2, 6, 3);
        particle2 = new Particle(gameState, 0, this.posX + 10, this.posY - 10, 2, -2, 1, 2, 8, 3);
        particle3 = new Particle(gameState, 0, this.posX - 3, this.posY + 3, 2, 2, 1, 2, 2, 3);
        particle4 = new Particle(gameState, 0, this.posX + 10, this.posY + 3, -2, 2, 1, 2, 4, 3);
        gameState.getParticles().add(particle1);
        gameState.getParticles().add(particle2);
        gameState.getParticles().add(particle3);
        gameState.getParticles().add(particle4);
        particle1.spawn(5);
        particle2.spawn(5);
        particle3.spawn(5);
        particle4.spawn(5);
    }

    public boolean getActiveState() {
        return this.isActive;
    }
    public void setActiveState(boolean activeState) {
        this.isActive = activeState;
    }

    public boolean getInvincibilityState() {
        return this.isInvincible;
    }
    public void setInvincibilityState(boolean invincibilityState) {
        this.isInvincible = invincibilityState;
    }

}
