/*
 * Copyright 2025 Naomi (boyfailure.dev).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dev.boyfailure.vectorbreakout;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import javax.swing.JComponent;
import java.awt.geom.Line2D;

/**
 * The logic to draw the game on the screen.
 * @author Naomi (boyfailure.dev)
 * @since 20250323
 */
public class GameDisplay extends JComponent {

    // Change this to 1 if you build for older versions of Java. Older versions
    // will render a leading zero otherwise. (tested on JDK 1.5)
    byte iStart = 0;

    float beamX = 0; // Horizontal location of "vector beam"
    float beamY = 0; // Vertical location of "vector beam"

    public float beamScaleX = 1; // Horizontal drawing scale
    public float beamScaleY = 1; // Vertical drawing scale
    public float beamThicknessScale = 1; // Thickness of lines

    private GameState gameState;

    public GameDisplay(GameState gameState) {
        this.gameState = gameState;
    }

    public float getBeamScaleX() {
        return this.beamScaleX;
    }
    public float getBeamScaleY() {
        return this.beamScaleY;
    }
    public float getBeamThicknessScale() {
        return this.beamThicknessScale;
    }
    public void setBeamScaleX(float beamScale) {
        this.beamScaleX = beamScale;
    }
    public void setBeamScaleY(float beamScale) {
        this.beamScaleY = beamScale;
    }
    public void setBeamThicknessScale(float beamScale) {
        this.beamThicknessScale = beamScale;
    }

    public synchronized void paint(Graphics gameGraphics) {

        super.paint(gameGraphics);

        Graphics2D g2 = (Graphics2D) gameGraphics;

        // Walls
        if (gameState.getGameStartedState()) {
            g2.setColor(Color.gray);
            g2.setStroke(bt(10));
            resetBeam();
            moveBeam(15, 0);
            g2.draw(drawVec(0, 600));
            moveBeam(770, 0);
            g2.draw(drawVec(0, -600));
            moveBeam(15, -5);
            g2.draw(drawVec(-800, 0));
        }

        // Bricks
        if (gameState.getCurrentLevel() != null) {
            if (gameState.getCurrentLevel().brickList != null) {
                for (Brick brick : gameState.getCurrentLevel().brickList) {
                    if (!brick.getBrokenState()) {
                        resetBeam();
                        g2.setStroke(bt(3));
                        moveBeam(brick.posX + 1, brick.posY + 1);
                        g2.setColor(vc(brick.brickColor));
                        g2.draw(drawVec(brick.width - 2, 0));
                        g2.draw(drawVec(0, -(brick.height - 2)));
                        g2.draw(drawVec(-(brick.width - 2), 0));
                        g2.draw(drawVec(0, brick.height - 2));
                    }
                }
            }
        }

        // Paddles
        g2.setColor(Color.white);
        for (Paddle paddle : gameState.getPaddles()) {
            if (paddle.getActiveState()) {
                resetBeam();
                g2.setStroke(bt(3));
                moveBeam(paddle.posX, paddle.posY);
                g2.draw(drawVec(paddle.width, 0));
                g2.draw(drawVec(0, -(paddle.height)));
                g2.draw(drawVec(-(paddle.width), 0));
                g2.draw(drawVec(0, paddle.height));
            }
        }

        // Balls
        g2.setStroke(bt(3));
        g2.setColor(Color.yellow);
        for (Ball ball : gameState.getBalls()) {
            if (ball.getActiveState()) {
                resetBeam();
                moveBeam(ball.posX, ball.posY);
                g2.draw(drawVec(ball.width, 0));
                g2.draw(drawVec(0, -ball.height));
                g2.draw(drawVec(-ball.width, 0));
                g2.draw(drawVec(0, ball.height));
            }
        }

        // Particles
        for (Particle particle : gameState.getParticles()) {
            if (particle.getActiveState()) {
                resetBeam();
                moveBeam(particle.particleX, particle.particleY);
                g2.setStroke(bt(particle.vectorScale));
                g2.setColor(vc(particle.particleColor));
                g2.draw(drawVec(particle.particleX2, particle.particleY2));
            }
        }

        /*
         * TextElements
         *
         * Note on rendering numbers:
         *
         * I am aware that I could be using a switch statement for these.
         * However, this did not compile when testing in JDK 1.5, as it would
         * refuse to accept anything other than numbers in a switch statement.
         * Since I am a silly little nerd who wishes to use this on my older
         * systems, I used these goofy if/else statements so I could compile
         * on the older JDKs and play on my older systems...
         */
        for (TextElement element : gameState.getTextElements()) {
            if (element.getActiveState() && element.getVisibleState()) {
                if (element.movesUpOnFrame()) {element.moveTo(element.getPositionX(), element.getPositionY() - 1);}
                resetBeam();
                moveBeam(element.getPositionX(), element.getPositionY());

                // Save the current beam scale to revert once TextElement is done rendering
                float oldBeamScaleX = this.beamScaleX;
                float oldBeamScaleY = this.beamScaleY;

                // Adjust the beam scale to accomodate the scale of the TextElement
                this.beamScaleX *= element.getTextScale();
                this.beamScaleY *= element.getTextScale();

                // Render the TextElement
                g2.setColor(vc(element.getColor()));
                g2.setStroke(bt(element.getVectorThickness()));
                String[] stringToRender = element.getText().split("");

                //<editor-fold defaultstate="collapsed" desc=" Beam routines for letters/numbers ">
                for (short i = iStart; i < stringToRender.length; i++) {
                    if ("1".equals(stringToRender[i])) {
                        moveBeam(7, 0);
                        g2.draw(drawVec(0, -24));
                        moveBeam(11, 24);
                    }
                    else if ("2".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(-14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 0);
                    }
                    else if ("3".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(-14, 0));
                        moveBeam(14, 0);
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(-14, 0));
                        moveBeam(18, 0);
                    }
                    else if ("4".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(14, 0));
                        moveBeam(0, -12);
                        g2.draw(drawVec(0, 24));
                        moveBeam(4, 0);
                    }
                    else if ("5".equals(stringToRender[i]) || "S".equals(stringToRender[i])) {
                        moveBeam(14, -24);
                        g2.draw(drawVec(-14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(-14, 0));
                        moveBeam(18, 0);
                    }
                    else if ("6".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(0, 24));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, -12));
                        g2.draw(drawVec(-14, 0));
                        moveBeam(18, 12);
                    }
                    else if ("7".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 24));
                        moveBeam(4, 0);
                    }
                    else if ("8".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 24));
                        g2.draw(drawVec(-14, 0));
                        moveBeam(0, -12);
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 12);
                    }
                    else if ("9".equals(stringToRender[i])) {
                        moveBeam(14, 0);
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(-14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 12);
                    }
                    else if ("0".equals(stringToRender[i]) || "O".equals(stringToRender[i])) {
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(-14, 0));
                        g2.draw(drawVec(0, 24));
                        moveBeam(18, 0);
                    }
                    // Simulates a newline in a TextElement
                    else if ("\\".equals(stringToRender[i])) {
                        moveBeam(-(beamX - element.getPositionX()), 30);
                    }
                    else if ("A".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -17));
                        g2.draw(drawVec(7, -7));
                        g2.draw(drawVec(7, 7));
                        g2.draw(drawVec(0, 17));
                        moveBeam(-14, -10);
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 10);
                    }
                    else if ("B".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(11, 0));
                        g2.draw(drawVec(3, 3));
                        g2.draw(drawVec(0, 6));
                        g2.draw(drawVec(-3, 3));
                        g2.draw(drawVec(3, 3));
                        g2.draw(drawVec(0, 6));
                        g2.draw(drawVec(-3, 3));
                        g2.draw(drawVec(-11, 0));
                        moveBeam(0, -12);
                        g2.draw(drawVec(11, 0));
                        moveBeam(7, 12);
                    }
                    else if ("C".equals(stringToRender[i])) {
                        moveBeam(14, -24);
                        g2.draw(drawVec(-14, 0));
                        g2.draw(drawVec(0, 24));
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 0);
                    }
                    else if ("D".equals(stringToRender[i])) {
                        g2.draw(drawVec(7, 0));
                        g2.draw(drawVec(7, -7));
                        g2.draw(drawVec(0, -10));
                        g2.draw(drawVec(-7, -7));
                        g2.draw(drawVec(-7, 0));
                        g2.draw(drawVec(0, 24));
                        moveBeam(18, 0);
                    }
                    else if ("E".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(14, 0));
                        moveBeam(-14, 12);
                        g2.draw(drawVec(14, 0));
                        moveBeam(-14, 12);
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 0);
                    }
                    else if ("F".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(14, 0));
                        moveBeam(-14, 12);
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 12);
                    }
                    else if ("G".equals(stringToRender[i])) {
                        moveBeam(14, -17);
                        g2.draw(drawVec(0, -7));
                        g2.draw(drawVec(-14, 0));
                        g2.draw(drawVec(0, 24));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, -8));
                        g2.draw(drawVec(-8, 0));
                        moveBeam(12, 8);
                    }
                    else if ("H".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        moveBeam(14, 0);
                        g2.draw(drawVec(0, 24));
                        moveBeam(-14, -12);
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 12);
                    }
                    else if ("I".equals(stringToRender[i])) {
                        g2.draw(drawVec(14, 0));
                        moveBeam(-14, -24);
                        g2.draw(drawVec(14, 0));
                        moveBeam(-7, 0);
                        g2.draw(drawVec(0, 24));
                        moveBeam(11, 0);
                    }
                    else if ("J".equals(stringToRender[i])) {
                        moveBeam(0, -7);
                        g2.draw(drawVec(7, 7));
                        g2.draw(drawVec(7, 0));
                        g2.draw(drawVec(0, -24));
                        moveBeam(4, 24);
                    }
                    else if ("K".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        moveBeam(14, 0);
                        g2.draw(drawVec(-14, 12));
                        g2.draw(drawVec(14, 12));
                        moveBeam(4, 0);
                    }
                    else if ("L".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(0, 24));
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 0);
                    }
                    else if ("M".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(7, 7));
                        g2.draw(drawVec(7, -7));
                        g2.draw(drawVec(0, 24));
                        moveBeam(4, 0);
                    }
                    else if ("N".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(14, 24));
                        g2.draw(drawVec(0, -24));
                        moveBeam(4, 24);
                    }
                    else if ("P".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(-14, 0));
                        moveBeam(18, 12);
                    }
                    else if ("Q".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 18));
                        g2.draw(drawVec(-6, 6));
                        g2.draw(drawVec(-8, 0));
                        moveBeam(8, -6);
                        g2.draw(drawVec(6, 6));
                        moveBeam(4, 0);
                    }
                    else if ("R".equals(stringToRender[i])) {
                        g2.draw(drawVec(0, -24));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, 12));
                        g2.draw(drawVec(-14, 0));
                        moveBeam(2, 0);
                        g2.draw(drawVec(12, 12));
                        moveBeam(4, 0);
                    }
                    else if ("T".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(14, 0));
                        moveBeam(-7, 0);
                        g2.draw(drawVec(0, 24));
                        moveBeam(11, 0);
                    }
                    else if ("U".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(0, 24));
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(0, -24));
                        moveBeam(4, 24);
                    }
                    else if ("V".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(7, 24));
                        g2.draw(drawVec(7, -24));
                        moveBeam(4, 24);
                    }
                    else if ("W".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(0, 24));
                        g2.draw(drawVec(7, -7));
                        g2.draw(drawVec(7, 7));
                        g2.draw(drawVec(0, -24));
                        moveBeam(4, 24);
                    }
                    else if ("X".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(14, 24));
                        moveBeam(-14, 0);
                        g2.draw(drawVec(14, -24));
                        moveBeam(4, 24);
                    }
                    else if ("Y".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(7, 7));
                        g2.draw(drawVec(7, -7));
                        moveBeam(-7, 7);
                        g2.draw(drawVec(0, 17));
                        moveBeam(11, 0);
                    }
                    else if ("Z".equals(stringToRender[i])) {
                        moveBeam(0, -24);
                        g2.draw(drawVec(14, 0));
                        g2.draw(drawVec(-14, 24));
                        g2.draw(drawVec(14, 0));
                        moveBeam(4, 0);
                    }
                    else if (".".equals(stringToRender[i])) {
                        moveBeam(7, 0);
                        g2.draw(drawVec(0, -2));
                        moveBeam(11, 2);
                    }
                    else if (",".equals(stringToRender[i])) {
                        moveBeam(7, 0);
                        g2.draw(drawVec(0, -5));
                        moveBeam(11, 5);
                    }
                    else if (":".equals(stringToRender[i])) {
                        moveBeam(7, 0);
                        g2.draw(drawVec(0, -2));
                        moveBeam(0, -10);
                        g2.draw(drawVec(0, -2));
                        moveBeam(11, 14);
                    }
                    else if (";".equals(stringToRender[i])) {
                        moveBeam(7, 0);
                        g2.draw(drawVec(0, -5));
                        moveBeam(0, -7);
                        g2.draw(drawVec(0, -2));
                        moveBeam(11, 14);
                    }
                    else if ("!".equals(stringToRender[i])) {
                        moveBeam(7, 0);
                        g2.draw(drawVec(0, -2));
                        moveBeam(0, -10);
                        g2.draw(drawVec(0, -12));
                        moveBeam(11, 24);
                    }
                    else if ("(".equals(stringToRender[i])) {
                        moveBeam(10, 0);
                        g2.draw(drawVec(-6, -6));
                        g2.draw(drawVec(0, -12));
                        g2.draw(drawVec(6, -6));
                        moveBeam(8, 24);
                    }
                    else if (")".equals(stringToRender[i])) {
                        moveBeam(10, 0);
                        g2.draw(drawVec(6, -6));
                        g2.draw(drawVec(0, -12));
                        g2.draw(drawVec(-6, -6));
                        moveBeam(8, 24);
                    }
                    else { // Blank space
                        moveBeam(18, 0);
                    }
                }
                //</editor-fold>

                // Revert back to the previous beam scale
                this.beamScaleX = oldBeamScaleX;
                this.beamScaleY = oldBeamScaleY;
            }
        }

        gameState.incrementActualFrames(); // Count up actual frames rendered

        // Debug Menu
        if (gameState.getDebugMenuState()) {
            g2.setColor(Color.white);
            g2.drawString(gameState.getGameName() + " v" + gameState.getGameVersion(), 1, 12);
            g2.drawString("Frame: " + gameState.getFrameCounter(), 1, 24);
            g2.drawString("Horizontal beamscale: " + beamScaleX, 1, 36);
            g2.drawString("Vertical beamscale: " + beamScaleY, 1, 48);
            g2.drawString("Paused: " + gameState.getPausedState(), 1, 60);
            g2.drawString("Target framerate: " + gameState.getTargetFrameRate(), 150, 24);
            g2.drawString("Actual framerate: " + gameState.getActualFrameRate(), 150, 36);
            g2.drawString("actualFrames: " + gameState.getActualFrames(), 150, 48);
            g2.drawString("LFSR: " + gameState.getPokey().getDecimalValue(), 150, 60);
            g2.drawString("Ball count: " + gameState.getBalls().size(), 300, 24);
            g2.drawString("Particle count: " + gameState.getParticles().size(), 300, 36);
            g2.drawString("Paddle count: " + gameState.getPaddles().size(), 300, 48);
            g2.drawString("Text count: " + gameState.getTextElements().size(), 300, 60);
            g2.drawString("Brick count: " + gameState.getCurrentLevel().brickList.size(), 300, 72);
            short padBoxX, padBoxY, padBoxX2, padBoxY2;
            g2.setStroke(new BasicStroke(1));

            // Paddles
            g2.setColor(vc(10, 0.5f));
            for (Paddle paddle : gameState.getPaddles()) {
                padBoxX = (short) paddle.hitBox.getBounds().getBounds2D().getMinX();
                padBoxY = (short) paddle.hitBox.getBounds().getBounds2D().getMinY();
                padBoxX2 = (short) paddle.hitBox.getBounds().getBounds2D().getWidth();
                padBoxY2 = (short) paddle.hitBox.getBounds().getBounds2D().getHeight();
                g2.fillRect(padBoxX, padBoxY + (40 - paddle.height), padBoxX2, padBoxY2);
                padBoxX = (short) paddle.hitBoxLeft.getBounds().getBounds2D().getMinX();
                padBoxY = (short) paddle.hitBoxLeft.getBounds().getBounds2D().getMinY();
                padBoxX2 = (short) paddle.hitBoxLeft.getBounds().getBounds2D().getWidth();
                padBoxY2 = (short) paddle.hitBoxLeft.getBounds().getBounds2D().getHeight();
                g2.fillRect(padBoxX, padBoxY + (40 - paddle.height), padBoxX2, padBoxY2);
                padBoxX = (short) paddle.hitBoxRight.getBounds().getBounds2D().getMinX();
                padBoxY = (short) paddle.hitBoxRight.getBounds().getBounds2D().getMinY();
                padBoxX2 = (short) paddle.hitBoxRight.getBounds().getBounds2D().getWidth();
                padBoxY2 = (short) paddle.hitBoxRight.getBounds().getBounds2D().getHeight();
                g2.fillRect(padBoxX, padBoxY + (40 - paddle.height), padBoxX2, padBoxY2);
            }

            // Balls
            g2.setColor(vc(11, 1.0f));
            for (Ball ball : gameState.getBalls()) {
                padBoxX = (short) ball.hitBox.getBounds().getBounds2D().getMinX();
                padBoxY = (short) ball.hitBox.getBounds().getBounds2D().getMinY();
                padBoxX2 = (short) ball.hitBox.getBounds().getBounds2D().getWidth();
                padBoxY2 = (short) ball.hitBox.getBounds().getBounds2D().getHeight();
                g2.fillRect(padBoxX, padBoxY + (40 - ball.height), padBoxX2, padBoxY2);
            }

            // Bricks
            g2.setColor(vc(12, 0.5f));
            for (Brick brick : gameState.getCurrentLevel().brickList) {
                padBoxX = (short) brick.hitBox.getBounds().getBounds2D().getMinX();
                padBoxY = (short) brick.hitBox.getBounds().getBounds2D().getMinY();
                padBoxX2 = (short) brick.hitBox.getBounds().getBounds2D().getWidth();
                padBoxY2 = (short) brick.hitBox.getBounds().getBounds2D().getHeight();
                g2.fillRect(padBoxX, padBoxY + (40 - brick.height), padBoxX2, padBoxY2);
                padBoxX = (short) brick.hitBoxLeft.getBounds().getBounds2D().getMinX();
                padBoxY = (short) brick.hitBoxLeft.getBounds().getBounds2D().getMinY();
                padBoxX2 = (short) brick.hitBoxLeft.getBounds().getBounds2D().getWidth();
                padBoxY2 = (short) brick.hitBoxLeft.getBounds().getBounds2D().getHeight();
                g2.fillRect(padBoxX, padBoxY + (40 - brick.height), padBoxX2, padBoxY2);
                padBoxX = (short) brick.hitBoxRight.getBounds().getBounds2D().getMinX();
                padBoxY = (short) brick.hitBoxRight.getBounds().getBounds2D().getMinY();
                padBoxX2 = (short) brick.hitBoxRight.getBounds().getBounds2D().getWidth();
                padBoxY2 = (short) brick.hitBoxRight.getBounds().getBounds2D().getHeight();
                g2.fillRect(padBoxX, padBoxY + (40 - brick.height), padBoxX2, padBoxY2);
            }
        }
        
    }

    public Line2D.Float drawVec(float x, float y) {
        float oldBeamX = beamX;
        float oldBeamY = beamY;
        beamX += (x * beamScaleX);
        beamY += (y * beamScaleY);
        return new Line2D.Float(oldBeamX, oldBeamY, beamX, beamY);
    }

    public Color vc(int colorID) {
        return vc(colorID, 1.0f);
    }

    public Color vc(int colorID, float alpha) {
        switch ((byte) colorID) {
            case 0: return new Color(0, 0, 0, alpha); // black
            // default is white but will be written elsewhere as 1
            case 2: return new Color(0.66f, 0.66f, 0.66f, alpha); // grey
            case 3: return new Color(0.5f, 0.5f, 0.5f, alpha); // dark grey
            case 4: return new Color(0.66f, 0, 0, alpha); // dark red
            case 5: return new Color(0, 0.66f, 0, alpha); // dark green
            case 6: return new Color(0, 0, 0.66f, alpha); // dark blue
            case 7: return new Color(1, 0, 0, alpha); // red
            case 8: return new Color(0, 1, 0, alpha); // green
            case 9: return new Color(0, 0, 1, alpha); // blue
            case 10: return new Color(1, 0.5f, 0.5f, alpha); // light red
            case 11: return new Color(0.5f, 1, 0.5f, alpha); // light green
            case 12: return new Color(0.5f, 0.5f, 1, alpha); // light blue
            case 13: return new Color(1, 0.66f, 0, alpha); // orange
            case 14: return new Color(1, 1, 0, alpha); // yellow
            case 15: return new Color(0, 0.66f, 0.66f, alpha); // cyan
            case 16: return new Color(0, 1, 1, alpha); // aqua
            case 17: return new Color(0.66f, 0, 0.66f, alpha); // purple
            case 18: return new Color(1, 0, 1, alpha); // magenta
            case 19: return new Color(1, 0.5f, 0.75f, alpha); // pastel pink
            case 20: return new Color(0.5f, 0.66f, 1, alpha); // pastel blue
            default: return new Color(1, 1, 1, alpha); // white
        }
    }

    public BasicStroke bt(float size) {
        return new BasicStroke(size * beamThicknessScale);
    }

    public void moveBeam(float x, float y) {
        beamX += (x * beamScaleX);
        beamY += (y * beamScaleY);
    }

    public void resetBeam() {
        beamX = 0;
        beamY = 40 * beamScaleY;
    }

}
