/*
 * Copyright 2025 Naomi (boyfailure.dev).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dev.boyfailure.vectorbreakout;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;

/**
 * The embeddable game object which contains the main loop and logic.
 * @author Naomi (boyfailure.dev)
 * @since 20250511
 */
public class GameState {

    /** The name of the game. */
    private String gameName = "Vector Breakout";
    /** The version of the game. */
    private String gameVersion = "1.0";
    /** Determines whether the game has been started. */
    private boolean isGameStarted = false;
    /** Determines whether the game has been paused. */
    private boolean isPaused = false;
    /** Determines whether the debug menu has been enabled. */
    private boolean debugMenuEnabled = false;
    /** Determines whether the game has been lost. */
    private boolean isLost = false;
    /** Determines whether the player is moving left. */
    private boolean movingLeft = false;
    /** Determines whether the player is moving right. */
    private boolean movingRight = false;
    /** Determines whether "Confetti Mode" (long-lasting particles) has been enabled. */
    private boolean confettiMode = false;
    /** Determines whether the collector function is running. */
    private boolean collectorActive = false;
    /** The number of lives the player has. */
    private byte lives = 5;
    /** The target frame rate for the game. */
    private short targetFrameRate = 60;
    private short actualFrames = 0;
    private short actualFrameRate = 0;
    private byte gameTickRate = 60;
    private short internalResX = 800;
    private short internalResY = 600;
    private short windowResX = 1280;
    private short windowResY = 960;
    private short level = 1;
    private short bricksOnScreen = 0;
    private int score = 0;
    private long frameCounter = 0;

    private Level currentLevel;

    private ArrayList<Ball> ballList;
    private ArrayList<Particle> particleList;
    private ArrayList<Paddle> paddleList;
    private ArrayList<TextElement> textElementList;

    private ArrayList<Ball> ballCollector;
    private ArrayList<Particle> particleCollector;
    private ArrayList<Paddle> paddleCollector;
    private ArrayList<TextElement> textElementCollector;
    private ArrayList<Brick> brickCollector;

    private LFSR pokey;

    public GameState() {
        this.initComponents();
    }

    //<editor-fold defaultstate="collapsed" desc=" Event listeners ">

    public KeyListener getKeyListener() {
        return new GameKeyListener();
    }
    
    public ActionListener getGameStateUpdateActionListener() {
        return new GameStateUpdateListener();
    }
    
    public ActionListener getGameTickActionListener() {
        return new GameTickActionListener();
    }

    class GameTickActionListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            onGameTick();
        }
    }

    class GameKeyListener implements KeyListener {
        public void keyPressed(KeyEvent e) {
            switch(e.getKeyCode()) {
                case 27: // Escape - Toggle debug menu
                    toggleDebugMenu();
                    break;
                case 10: // Enter - Pause/unpause
                    if (GameState.this.getGameStartedState()) {GameState.this.togglePausedState();}
                    break;
                case 35: // End - Start debug level
                    if (isLost || !GameState.this.getGameStartedState()) {GameState.this.setGameStartedState(true);GameState.this.newGame(true);}
                    break;
                case 32: // Space - Start normal game, spawn ball
                    if (isLost || !GameState.this.getGameStartedState()) {GameState.this.setGameStartedState(true);GameState.this.newGame();}
                    else if (!GameState.this.getPausedState()) {
                        if (GameState.this.getBalls().size() < 1) {
                            Ball newBall = new Ball(GameState.this, 200, 300, 7, 7, 0, true);
                            GameState.this.ballList.add(newBall);
                            newBall.spawnBall(true);
                        }
                    }
                    break;
                case 37: // Left - move left
                    if (GameState.this.getGameStartedState()) {movingLeft = true;}
                    break;
                case 39: // Right - move right
                    if (GameState.this.getGameStartedState()) {movingRight = true;}
                    break;
            }
        }
        public void keyReleased(KeyEvent e) {
            switch(e.getKeyCode()) {
                case 37: // Left
                    movingLeft = false;
                    break;
                case 39: // Right
                    movingRight = false;
                    break;
            }
        }
        public void keyTyped(KeyEvent e) {}
    }

    class GameStateUpdateListener implements ActionListener {
        public synchronized void actionPerformed(ActionEvent e) {

            GameState.this.collectorActive = true;
            GameState.this.ballCollector = new ArrayList<>();
            GameState.this.particleCollector = new ArrayList<>();
            GameState.this.paddleCollector = new ArrayList<>();
            GameState.this.textElementCollector = new ArrayList<>();
            GameState.this.brickCollector = new ArrayList<>();

            bricksOnScreen = 0;
            actualFrameRate = actualFrames;
            actualFrames = 0;
            if (GameState.this.currentLevel != null) {
                if (GameState.this.currentLevel.brickList != null) {
                    for (Brick brick : GameState.this.currentLevel.brickList) {
                        if (brick.getBrokenState()) {
                            GameState.this.brickCollector.add(brick);
                        }
                        else {
                            bricksOnScreen++;
                        }
                    }
                }
            }

            for (Ball ball : GameState.this.ballList) {
                if (!ball.getActiveState()) {
                    GameState.this.ballCollector.add(ball);
                }
            }

            for (Particle particle: GameState.this.particleList) {
                if (!particle.getActiveState()) {
                    GameState.this.particleCollector.add(particle);
                }
            }

            if (bricksOnScreen <= 0 && GameState.this.getGameStartedState() && !isLost) {
                level++;
                GameState.this.currentLevel = new Level(GameState.this, level);
                GameState.this.currentLevel.constructLevel();
            }

            GameState.this.ballList.removeAll(GameState.this.ballCollector);
            GameState.this.paddleList.removeAll(GameState.this.paddleCollector);
            GameState.this.particleList.removeAll(GameState.this.particleCollector);
            GameState.this.textElementList.removeAll(GameState.this.textElementCollector);
            if (GameState.this.currentLevel != null) {
                GameState.this.currentLevel.brickList.removeAll(GameState.this.brickCollector);
            }
            GameState.this.collectorActive = false;

        }
    }

    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Accessor methods (get/set) ">

    //<editor-fold defaultstate="collapsed" desc=" Version information accessors ">
    /**
     * Gets the name of the game.
     * @return the name of the game
     */
    public String getGameName() {
        return this.gameName;
    }
    /**
     * Gets the version number of the game.
     * @return the version number of the game
     */
    public String getGameVersion() {
        return this.gameVersion;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Pokey accessors ">
    /**
     * Gets the current instance of "Pokey", the game's LFSR.
     * @return pokey
     */
    public LFSR getPokey() {
        return this.pokey;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Tick rate accessors ">
    /**
     * Gets the current game tick rate, in Hz.
     * @return the current game tick rate (in Hz)
     */
    public byte getGameTickRate() {
        return this.gameTickRate;
    }

    /**
     * Sets the current game tick rate, in Hz.
     * @param newTickRate the game tick rate, in Hz
     */
    public void setGameTickRate(byte newTickRate) {
        this.gameTickRate = (byte) newTickRate;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Frame rate accessors ">
    /**
     * Gets the target game frame rate, in Hz.
     * @return the target game frame rate (in Hz)
     */
    public short getTargetFrameRate() {
        return this.targetFrameRate;
    }

    /**
     * Gets the actual game frame rate, in Hz.
     * @return the actual game frame rate (in Hz)
     */
    public short getActualFrameRate() {
        return this.actualFrameRate;
    }

    /**
     * Increments the frame counter.
     */
    public void incrementFrameCounter() {
        this.frameCounter++;
    }

    /**
     * Gets the value of the frame counter.
     * @return the value of the frame counter
     */
    public long getFrameCounter() {
        return this.frameCounter;
    }

    /**
     * Gets the actual frames drawn during the last second.
     * @return the actual frames drawn during the last second.
     */
    public short getActualFrames() {
        return this.actualFrames;
    }
    public void incrementActualFrames() {
        this.actualFrames++;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Pause state accessors ">
    public boolean getPausedState() {
        return this.isPaused;
    }
    public void togglePausedState() {
        this.isPaused = !this.isPaused;
        if (this.getPausedState()) {this.textElementList.get(6).getVisibleState(true);}
        else {this.textElementList.get(6).getVisibleState(false);}
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Debug menu state accessors ">
    public boolean getDebugMenuState() {
        return this.debugMenuEnabled;
    }
    public void toggleDebugMenu() {
        this.debugMenuEnabled = !this.debugMenuEnabled;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Game start state accessors ">
    public boolean getGameStartedState() {
        return this.isGameStarted;
    }
    public void setGameStartedState(boolean startState) {
        this.isGameStarted = startState;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Resolution accessors ">
    public short getInternalResX() {
        return this.internalResX;
    }
    public short getInternalResY() {
        return this.internalResY;
    }
    public short getWindowResX() {
        return this.windowResX;
    }
    public short getWindowResY() {
        return this.windowResY;
    }
    public void setWindowResX(short resolution) {
        this.windowResX = resolution;
    }
    public void setWindowResY(short resolution) {
        this.windowResY = resolution;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Lives accessors ">
    public byte getLives() {
        return this.lives;
    }
    public void addLives(int livesAdded) {
        if (livesAdded >= 127) {livesAdded = 127;}
        else if (livesAdded < 0) {livesAdded = 0;}
        this.lives += (byte) livesAdded;
    }
    public void addLife() {
        this.addLives(1);
    }
    public void removeLife() {
        this.lives--;
        if (this.lives < 0) {this.lives = 0;}
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Score accessors ">
    public int getScore() {
        return this.score;
    }
    public void addScore(int scoreAdded) {
        this.score += scoreAdded;
    }
    public void resetScore() {
        this.score = 0;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Level accessors ">
    public short getCurrentLevelID() {
        return this.level;
    }
    public void setCurrentLevelID(short levelID) {
        this.level = levelID;
    }
    public Level getCurrentLevel() {
        return this.currentLevel;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" List accessors ">
    public ArrayList<Ball> getBalls() {
        return this.ballList;
    }
    public ArrayList<Particle> getParticles() {
        return this.particleList;
    }
    public ArrayList<Paddle> getPaddles() {
        return this.paddleList;
    }
    public ArrayList<TextElement> getTextElements() {
        return this.textElementList;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Collector accessors ">
    public boolean getCollectorActiveState() {
        return this.collectorActive;
    }
    //</editor-fold>

    //</editor-fold>

    public synchronized void onGameTick() {

        if (this.getPausedState()) {return;}

        // Move player if keys are held
        try {
            for (Paddle paddle : this.paddleList) {
                if (this.movingLeft) {
                    paddle.movePaddle(false);
                }
                else if (this.movingRight) {
                    paddle.movePaddle(true);
                }
                else {
                    paddle.paddleSpeed = 0;
                }
            }
        } catch(java.util.ConcurrentModificationException except) {
            System.err.println("ConcurrentModificationException, onGameTick() Paddle iteration " + System.currentTimeMillis());
        }

        // Ball logic (movement, collision checks)
        try {
            for (Ball ball : this.ballList) {
                if (ball.getActiveState()) {
                    ball.moveBall();
                }
            }
        } catch(java.util.ConcurrentModificationException except) {
            System.err.println("ConcurrentModificationException, onGameTick() Ball iteration " + System.currentTimeMillis());
        }

        // Particles
        try {
            for (Particle particle : this.particleList) {
                if (particle.getActiveState()) {
                    particle.update();
                }
            }
        } catch(java.util.ConcurrentModificationException except) {
            System.err.println("ConcurrentModificationException, onGameTick() Particle iteration " + System.currentTimeMillis());
        }
    }

    public void showTitleScreen() {
        textElementList.get(0).activateTimer();
        textElementList.get(1).activateTimer();
        textElementList.get(2).activateTimer();
    }

    public synchronized void gameLose() {
        if (this.currentLevel != null) {
            for (Brick brick : this.currentLevel.brickList) {
                brick.setBrokenState(true);
                this.brickCollector.add(brick);
            }
        }
        for (Ball ball : this.ballList) {
            ball.setActiveState(false);
            this.ballCollector.add(ball);
        }
        for (Paddle paddle : this.paddleList) {
            paddle.setActiveState(false);
            paddle.cullPaddle();
            this.paddleCollector.add(paddle);
        }
        this.isLost = true;
        this.textElementList.get(3).getVisibleState(false);
        this.textElementList.get(4).getVisibleState(false);
        this.textElementList.get(5).getVisibleState(false);
        this.textElementList.get(8).getVisibleState(true);
        this.textElementList.get(9).getVisibleState(true);
        this.getGameStateUpdateActionListener().actionPerformed(new ActionEvent(this, 0, ""));
    }

    public void newGame() {
        newGame(false);
    }

    public void newGame(boolean debugLevel) {
        lives = 5;
        isLost = false;
        score = 0;
        if (debugLevel) {level = 32767;}
        else {level = 1;}
        frameCounter = 0;
        movingLeft = false;
        movingRight = false;
        this.paddleList.add(new Paddle(this, 350, 500, 100, 15, 0));
        this.textElementList.get(0).getVisibleState(false);
        this.textElementList.get(1).getVisibleState(false);
        this.textElementList.get(2).getVisibleState(false);
        this.textElementList.get(3).getVisibleState(true);
        this.textElementList.get(4).getVisibleState(true);
        this.textElementList.get(5).getVisibleState(true);
        textElementList.get(3).setText(String.valueOf(this.score));
        textElementList.get(4).setText(String.valueOf(this.lives));
        textElementList.get(5).setText(String.valueOf(this.level));
        this.textElementList.get(8).getVisibleState(false);
        this.textElementList.get(9).getVisibleState(false);
        this.currentLevel = new Level(this, level);
        this.currentLevel.constructLevel();
        this.updateStaticTextElements();
    }

    public synchronized void initComponents() {
        this.ballList = new ArrayList<>();
        this.particleList = new ArrayList<>();
        this.paddleList = new ArrayList<>();
        this.textElementList = new ArrayList<>();

        this.pokey = new LFSR();

                                                                                                    // List of indices
        textElementList.add(new TextElement(this, "VECTOR BREAKOUT", 2.5f, 67, 280, 15, 2));        // 0 - Title
        textElementList.add(new TextElement(this, "MMXXV BOYFAILURE.DEV", 0.75f, 259, 315, 15, 1)); // 1 - Copyright
        textElementList.add(new TextElement(this, "PRESS SPACE TO BEGIN", 1, 215, 345, 15, 2));     // 2 - Space
        textElementList.add(new TextElement(this, String.valueOf(score), 1, 20, -13, 1, 2));        // 3 - Score
        textElementList.add(new TextElement(this, String.valueOf(lives), 1, 200, -13, 1, 2));       // 4 - Lives
        textElementList.add(new TextElement(this, String.valueOf(level), 1, 380, -13, 1, 2));       // 5 - Level
        textElementList.add(new TextElement(this, "PAUSED", 5, 142, 310, 1, 4));                    // 6 - Paused
        textElementList.add(new TextElement(this, "Level Name", 1, 24, 480, 15, 2, 3000));          // 7 - Level Name
        textElementList.add(new TextElement(this, "   GAME OVER   ", 2.5f, 67, 280, 10, 2));        // 8 - Lose
        textElementList.add(new TextElement(this, "PRESS SPACE TO RETRY", 1, 215, 315, 10, 1));     // 9 - Retry
    }
    
    public void updateStaticTextElements() {
        textElementList.get(3).setText(String.valueOf(score));
        textElementList.get(4).setText(String.valueOf(lives));
        textElementList.get(5).setText(String.valueOf(level));
    }

}
