/*
 * Copyright 2025 Naomi (boyfailure.dev).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dev.boyfailure.vectorbreakout;

import java.awt.Point;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Fields of text that display on screen.
 * @author Naomi (boyfailure.dev)
 * @since 20250326
 */
public class TextElement {

    private String text;
    private boolean active = true;
    private boolean isVisible = false;
    private boolean moveUpOnFrame = false;
    private float textScale = 1;
    private float vectorThickness = 1;
    private byte textColor;
    private short x;
    private short y;
    private short duration = 0;
    private long creationTime;

    private GameState gameState;

    /**
     * Constructs a text object to be displayed on screen.
     * @param gameState the GameState in which the TextElement will reside
     * @param text the text that will be displayed
     * @param textScale the size of the text on screen
     * @param x the x coordinate of the text's origin point
     * @param y the y coordinate of the text's origin point
     * @param textColor the color of the text
     * @param vectorThickness the thickness of the vector lines
     */
    public TextElement(GameState gameState, String text, float textScale, int x, int y, int textColor, float vectorThickness) {
        this.gameState = gameState;
        this.text = text;
        this.textScale = textScale;
        this.x = (short) x;
        this.y = (short) y;
        this.textColor = (byte) textColor;
        this.vectorThickness = vectorThickness;
    }
    
    /**
     * Constructs a text object that disappears after a set amount of time.
     * @param gameState the GameState in which the TextElement will reside
     * @param text the text that will be displayed
     * @param textScale the size of the text on screen
     * @param x the x coordinate of the text's origin point
     * @param y the y coordinate of the text's origin point
     * @param textColor the color of the text
     * @param vectorThickness the thickness of the vector lines
     * @param duration how long the text will stay on screen, in milliseconds
     */
    public TextElement(GameState gameState, String text, float textScale, int x, int y, int textColor, float vectorThickness, int duration) {
        this.gameState = gameState;
        this.text = text;
        this.textScale = textScale;
        this.x = (short) x;
        this.y = (short) y;
        this.textColor = (byte) textColor;
        this.vectorThickness = vectorThickness;
        this.duration = (short) duration;
        this.creationTime = System.currentTimeMillis();
    }

    //<editor-fold defaultstate="collapsed" desc=" Accessor methods (get/set) ">

    //<editor-fold defaultstate="collapsed" desc=" Active state accessors ">
    /**
     * Gets the active status of the TextElement.
     * @return the active status of the TextElement
     */
    public boolean getActiveState() {
        return this.active;
    }
    /**
     * Sets the active status of the TextElement.
     * @param activeState the active status of the TextElement
     */
    public void setActiveState(boolean activeState) {
        this.active = activeState;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Visibility state accessors ">
    /**
     * Gets the visibility state of the TextElement.
     * @return the visibility state of the TextElement
     */
    public boolean getVisibleState() {
        return this.isVisible;
    }
    /**
     * Sets the visible state of the TextElement.
     * @param visibleState the visible state of the TextElement
     */
    public void getVisibleState(boolean visibleState) {
        this.isVisible = visibleState;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Color code accessors ">
    /**
     * Gets the color code of the TextElement.
     * @return the color code of the TextElement
     */
    public byte getColor() {
        return this.textColor;
    }
    /**
     * Sets the color code of the TextElement.
     * @param textColor the color code of the TextElement
     */
    public void setColor(int textColor) {
        this.textColor = (byte) textColor;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Scaling accessors ">
    /**
     * Returns the thickness of the TextElement's vector lines.
     * @return the thickness of the TextElement's vector lines
     */
    public float getVectorThickness() {
        return this.vectorThickness;
    }
    /**
     * Sets the thickness of the TextElement's vector lines.
     * @param vectorThickness the new thickness of the TextElement's vector lines
     */
    public void setVectorThickness(float vectorThickness) {
        this.vectorThickness = vectorThickness;
    }
    /**
     * Returns the scale of the text.
     * @return the scale of the text
     */
    public float getTextScale() {
        return this.textScale;
    }
    /**
     * Sets the scale of the text.
     * @param textScale the new scale of the text
     */
    public void setTextScale(float textScale) {
        this.textScale = textScale;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Text value accessors ">
    /**
     * Gets the text in the TextElement.
     * @return the text in the TextElement
     */
    public String getText() {
        return this.text;
    }
    /**
     * Changes the text that the TextElement will display.
     * @param text the text that the TextElement will display
     */
    public void setText(String text) {
        this.text = text;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Moves up on frame accessors ">
    public boolean movesUpOnFrame() {
        return this.moveUpOnFrame;
    }
    public void moveUpOnFrame() {
        this.moveUpOnFrame = true;
    }
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc=" Position accessors ">
    public Point getPosition() {
        return new Point(this.x, this.y);
    }
    public short getPositionX() {
        return this.x;
    }
    public short getPositionY() {
        return this.y;
    }
    //</editor-fold>

    //</editor-fold>

    /**
     * Prepares the TextElement for rendering.
     */
    public void activateTimer() {
        this.setActiveState(true);
        long testLong = this.creationTime;
        if (this.duration > 0) {
            Timer levelTitleGTFO = new Timer();
            levelTitleGTFO.schedule(new TimerTask() {
                public void run() {
                    TextElement.this.getVisibleState(false);
                }
            }, this.duration);
        }
        this.getVisibleState(true);
    }

    /**
     * Prepares the TextElement for culling.
     */
    public void cull() {
        this.active = false;
        this.x = 32767;
        this.y = 32767;
    }

    /**
     * Changes the text that the TextElement will display.
     * @param x The x coordinate of the text's origin point.
     * @param y The y coordinate of the text's origin point.
     */
    public void moveTo(int x, int y) {
        this.x = (short) x;
        this.y = (short) y;
    }

}
